# =============================================================================
# OWSLib. Copyright (C) 2012 Jachym Cepicky
#
# Contact email: jachym.cepicky@gmail.com
#
# =============================================================================

import logging

from urllib.parse import urlencode
from owslib.crs import Crs
from owslib.util import Authentication, build_get_url
from owslib.feature.schema import get_schema
from owslib.feature.postrequest import PostRequest_1_1_0, PostRequest_2_0_0

LOGGER = logging.getLogger(__name__)


class WebFeatureService_(object):
    """Base class for WebFeatureService implementations"""

    def __init__(self, auth=None):
        self.auth = auth or Authentication()

    def getBBOXKVP(self, bbox, typename):
        """Formate bounding box for KVP request type (HTTP GET)

        @param bbox: (minx,miny,maxx,maxy[,srs])
        @type bbox: List
        @param typename:  feature name
        @type typename: String
        @returns: String properly formated according to version and
            coordinate reference system
        """
        srs = None

        # srs of the bbox is specified in the bbox as fifth paramter
        if len(bbox) == 5:
            srs = Crs(bbox[4])
        # take default srs
        else:
            srs = self.contents[typename[0]].crsOptions[0]

        # 1.1.0 and 2.0.0 have same encoding
        if self.version in ["1.1.0", "2.0.0"]:

            # format bbox parameter
            if srs.encoding == "urn":
                if srs.axisorder == "yx":
                    return "%s,%s,%s,%s,%s" % (
                        bbox[1],
                        bbox[0],
                        bbox[3],
                        bbox[2],
                        srs.getcodeurn(),
                    )
                else:
                    return "%s,%s,%s,%s,%s" % (
                        bbox[0],
                        bbox[1],
                        bbox[2],
                        bbox[3],
                        srs.getcodeurn(),
                    )
            else:
                return "%s,%s,%s,%s,%s" % (
                    bbox[0],
                    bbox[1],
                    bbox[2],
                    bbox[3],
                    srs.getcode(),
                )
        # 1.0.0
        else:
            return "%s,%s,%s,%s,%s" % (
                bbox[0],
                bbox[1],
                bbox[2],
                bbox[3],
                srs.getcode(),
            )

    def getBBOXPost(self, bbox, typename):
        """Format bounding box for Post requests

        @param bbox: (minx,miny,maxx,maxy[,srs])
        @type bbox: List
        @param typename:  feature name
        @type typename: String
        @returns: String properly formated according to version and
            coordinate reference system
        """
        srs = None

        # srs of the bbox is specified in the bbox as fifth paramter
        if len(bbox) == 5:
            srs = Crs(bbox[4])
        # take default srs
        else:
            srs = self.contents[typename[0]].crsOptions[0]

        formatted_bbox = [bbox[0], bbox[1], bbox[2], bbox[3]]
        if self.version in ["1.1.0", "2.0.0"]:
            if srs.axisorder == "yx" and srs.encoding == "urn":
                formatted_bbox = [bbox[1], bbox[0], bbox[3], bbox[2]]

            if self.version == "1.1.0":
                formatted_bbox.append(srs.getcodeurn())
                return formatted_bbox
            if self.version == "2.0.0":
                formatted_bbox.append(srs.getcodeuri1())
                return formatted_bbox
        else:
            formatted_bbox.append(srs.getcode())
            return formatted_bbox

    def create_post_request(self):
        """Creates an xml POST request according to WFS version."""

        if self.version in ['1.1.0']:
            return PostRequest_1_1_0()

        if self.version in ['2.0', '2.0.0']:
            return PostRequest_2_0_0()

    def getSRS(self, srsname, typename):
        """Returns None or Crs object for given name

        @param typename:  feature name
        @type typename: String
        """
        if not isinstance(srsname, Crs):
            srs = Crs(srsname)
        else:
            srs = srsname

        try:
            index = self.contents[typename].crsOptions.index(srs)
            # Return the Crs string that was pulled directly from the
            # GetCaps document (the 'id' attribute in the Crs object).
            return self.contents[typename].crsOptions[index]
        except ValueError:
            options = ", ".join([crs.id for crs in self.contents[typename].crsOptions])
            LOGGER.warning(
                "Requested srsName %r is not declared as being "
                "allowed for requested typename %r. "
                "Options are: %r.",
                srs.getcode(),
                typename,
                options,
            )
            return None

    def getGETGetFeatureRequest(
        self,
        typename=None,
        filter=None,
        bbox=None,
        featureid=None,
        featureversion=None,
        propertyname=None,
        maxfeatures=None,
        srsname=None,
        storedQueryID=None,
        storedQueryParams=None,
        outputFormat=None,
        method="Get",
        startindex=None,
        sortby=None,
    ):
        """Formulate proper GetFeature request using KVP encoding
        ----------
        typename : list
            List of typenames (string)
        filter : string
            XML-encoded OGC filter expression.
        bbox : tuple
            (left, bottom, right, top) in the feature type's coordinates == (minx, miny, maxx, maxy)
        featureid : list
            List of unique feature ids (string)
        featureversion : string
            Default is most recent feature version.
        propertyname : list
            List of feature property names. '*' matches all.
        maxfeatures : int
            Maximum number of features to be returned.
        srsname: string
            EPSG code to request the data in
        method : string
            Qualified name of the HTTP DCP method to use.
        outputFormat: string (optional)
            Requested response format of the request.
        startindex: int (optional)
            Start position to return feature set (paging in combination with maxfeatures)
        sortby: list (optional)
            List of property names whose values should be used to order
            (upon presentation) the set of feature instances that
            satify the query.

        There are 3 different modes of use

        1) typename and bbox (simple spatial query)
        2) typename and filter (==query) (more expressive)
        3) featureid (direct access to known features)
        """
        storedQueryParams = storedQueryParams or {}

        base_url = next(
            (
                m.get("url")
                for m in self.getOperationByName("GetFeature").methods
                if m.get("type").lower() == method.lower()
            )
        )

        request = {"service": "WFS", "version": self.version, "request": "GetFeature"}

        # check featureid
        if featureid:
            request["featureid"] = ",".join(featureid)
        elif bbox:
            request["bbox"] = self.getBBOXKVP(bbox, typename)
        elif filter:
            request["query"] = str(filter)
        if typename:
            typename = (
                [typename] if isinstance(typename, str) else typename
            )  # noqa: E721
            if int(self.version.split(".")[0]) >= 2:
                request["typenames"] = ",".join(typename)
            else:
                request["typename"] = ",".join(typename)
        if propertyname:
            request["propertyname"] = ",".join(propertyname)
        if sortby:
            request["sortby"] = ",".join(sortby)
        if featureversion:
            request["featureversion"] = str(featureversion)
        if maxfeatures:
            if int(self.version.split(".")[0]) >= 2:
                request["count"] = str(maxfeatures)
            else:
                request["maxfeatures"] = str(maxfeatures)
        if srsname:
            request["srsname"] = str(srsname)

            # Check if desired SRS is supported by the service for each
            # typename. Warning will be thrown if that SRS is not allowed.
            for name in typename:
                _ = self.getSRS(srsname, name)
        if startindex:
            request["startindex"] = str(startindex)
        if storedQueryID:
            request["storedQuery_id"] = str(storedQueryID)
            for param in storedQueryParams:
                request[param] = storedQueryParams[param]
        if outputFormat is not None:
            request["outputFormat"] = outputFormat

        return build_get_url(base_url, request)

    def getPOSTGetFeatureRequest(
        self,
        typename=None,
        filter=None,
        bbox=None,
        featureid=None,
        featureversion=None,
        propertyname=None,
        maxfeatures=None,
        storedQueryID=None,
        storedQueryParams=None,
        outputFormat=None,
        method="Post",
        startindex=None,
        sortby=None,
    ):
        """Formulate proper GetFeature request using KVP encoding
        ----------
        typename : list
            List of typenames (string)
        filter : string
            XML-encoded OGC filter expression.
        bbox : tuple
            (left, bottom, right, top) in the feature type's coordinates == (minx, miny, maxx, maxy)
        featureid : list
            List of unique feature ids (string)
        featureversion : string
            Default is most recent feature version.
        propertyname : list
            List of feature property names. Leave blank (None) to get all properties.
        maxfeatures : int
            Maximum number of features to be returned.
        method : string
            Qualified name of the HTTP DCP method to use.
        outputFormat: string (optional)
            Requested response format of the request.
        startindex: int (optional)
            Start position to return feature set (paging in combination with maxfeatures)
        storedQueryID : string
            A name identifying a prepared set available in WFS-service.
            WFS version 2.0.0 and above only.
        storedQueryParams : dict
            Variable amount of extra information sent to server related to
            storedQueryID to further define the requested data.
            WFS version 2.0.0 and above only.
            {'parameter_name': parameter_value}
        sortby: list (optional)
            List of property names whose values should be used to order
            (upon presentation) the set of feature instances that
            satify the query.

        There are 5 different modes of use

        1) typename and bbox (simple spatial query)
        2) typename and filter (==query) (more expressive)
        3) featureid (direct access to known features)
        4) storedQueryID and optional storedQueryParams
        5) filter only via Post method
        """

        try:
            base_url = next(
                (
                    m.get("url")
                    for m in self.getOperationByName("GetFeature").methods
                    if m.get("type").lower() == method.lower()
                )
            )
        except StopIteration:
            base_url = self.url

        if not typename and filter:
            return base_url, filter

        request = self.create_post_request()

        if storedQueryID:
            if self.version in ["1.0.0", "1.1.0"]:
                LOGGER.warning("Stored queries are only supported in version 2.0.0 and above.")
                return None

            storedQueryParams = storedQueryParams or {}
            request.create_storedquery(storedQueryID, storedQueryParams)
            data = request.to_string()
            return base_url, data

        typename = (
            [typename] if isinstance(typename, str) else typename
        )  # noqa: E721
        typenames = ",".join(typename)

        request.create_query(typenames)

        if featureid:
            featureid = (
                [featureid] if isinstance(featureid, str) else featureid
            )
            request.set_featureid(featureid)
        elif bbox:
            request.set_bbox(self.getBBOXPost(bbox, typename))
        elif filter is not None:
            request.set_filter(filter)

        if featureversion:
            request.set_featureversion(str(featureversion))
        if maxfeatures:
            request.set_maxfeatures(maxfeatures)
        if outputFormat:
            request.set_outputformat(outputFormat)
        if propertyname:
            propertyname = (
                [propertyname] if isinstance(propertyname, str) else propertyname
            )
            request.set_propertyname(propertyname)
        if sortby:
            sortby = (
                [sortby] if isinstance(sortby, str) else sortby
            )
            request.set_sortby(sortby)
        if startindex:
            request.set_startindex(startindex)

        data = request.to_string()
        return base_url, data

    def get_schema(self, typename):
        """
        Get layer schema compatible with :class:`fiona` schema object
        """
        return get_schema(self.url, typename, self.version, auth=self.auth)



# =============================================================================
# Copyright (c) 2022 Tom Kralidis
#
# Author: Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

from io import BytesIO
import logging
from typing import BinaryIO

from owslib.ogcapi import Collections
from owslib.util import Authentication

LOGGER = logging.getLogger(__name__)


class Maps(Collections):
    """Abstraction for OGC API - Maps"""

    def __init__(self, url: str, json_: str = None, timeout: int = 30,
                 headers: dict = None, auth: Authentication = None):
        __doc__ = Collections.__doc__  # noqa
        super().__init__(url, json_, timeout, headers, auth)

    def maps(self) -> list:
        """
        implements /collections filtered on maps

        @returns: `list` of filtered collections object
        """

        maps_ = []
        collections_ = super().collections()

        for c_ in collections_['collections']:
            for l_ in c_['links']:
                if 'map' in l_['rel']:
                    maps_.append(c_['id'])
                    break

        return maps_

    def map(self, collection_id: str, **kwargs: dict) -> BinaryIO:
        """
        implements /collection/{collectionId}/map

        @type collection_id: string
        @param collection_id: id of collection
        @type bbox: list
        @param bbox: list of minx,miny,maxx,maxy
        @type bbox_crs: str
        @param bbox_crs: CRS of output map
        @type datetime_: string
        @param datetime_: time extent or time instant
        @type width: int
        @param width: width of output map
        @type height: int
        @param height: height of output map
        @type transparent: bool
        @param transparent: whether to apply transparecy to
                            output map (default=True)
        @type style: str
        @param style: style name

        @returns: map image
        """

        kwargs_ = {}

        if 'bbox' in kwargs:
            kwargs_['bbox'] = ','.join(list(map(str, kwargs['bbox'])))
        if 'bbox_crs' in kwargs:
            kwargs_['bbox-crs'] = kwargs['bbox_crs']
        if 'datetime_' in kwargs:
            kwargs_['datetime'] = kwargs['datetime_']

        kwargs_['width'] = kwargs.get('width', 800)
        kwargs_['height'] = kwargs.get('height', 600)

        kwargs_['transparent'] = str(kwargs.get('transparent', True)).lower()

        if 'style' in kwargs:
            path = f'collections/{collection_id}/styles/{kwargs["style"]}/map'
        else:
            path = f'collections/{collection_id}/map'

        return BytesIO(self._request(path=path, as_dict=False, kwargs=kwargs_))



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2015 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

"""GM03 Core metadata parser http://www.geocat.ch/internet/geocat/en/home/documentation/gm03.html"""

from owslib import util
from owslib.etree import etree
from owslib.namespaces import Namespaces


# default variables
def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces(["gm03"])
    ns[None] = n.get_namespace("gm03")
    return ns


namespaces = get_namespaces()


class _GenericObject(object):
    """GM03 generic object type"""
    def __init__(self, md):
        """constructor"""

        self.tid = None

        if md is not None:
            self.tid = md.attrib.get('TID')


class _GenericObjectProperty(object):
    """GM03 generic object type"""
    def __init__(self, md):
        """constructor"""

        self.ref = None
        self.bid = None
        self.order_post = None

        if md is not None:
            self.ref = md.attrib.get('REF')
            self.bid = md.attrib.get('BID')
            self.order_pos = md.attrib.get('ORDER_POS')


class PT_Group(object):
    """PT_Group parser"""
    def __init__(self, md):
        """constructor"""

        self.language = util.testXMLValue(md.find(util.nspath_eval('gm03:language', namespaces)))
        self.country = util.testXMLValue(md.find(util.nspath_eval('gm03:country', namespaces)))
        self.character_set_code = util.testXMLValue(md.find(util.nspath_eval('gm03:characterSetCode', namespaces)))
        self.plain_text = util.testXMLValue(md.find(util.nspath_eval('gm03:plainText', namespaces)))
        self.plain_url = util.testXMLValue(md.find(util.nspath_eval('gm03:plainURL', namespaces)))


class PT_FreeText(object):
    """PT_FreeText parser"""
    def __init__(self, md):
        """constructor"""

        pt_groups = []
        for pt_group in md.findall(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.PT_FreeText/gm03:textGroup/gm03:GM03_2_1Core.Core.PT_Group', namespaces)):
            pt_groups.append(PT_Group(pt_group))

        self.pt_group = pt_groups


class PT_FreeURL(object):
    """PT_FreeURL parser"""
    def __init__(self, md):
        """constructor"""

        pt_groups = []
        for pt_group in md.findall(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.PT_FreeURL/gm03:URLGroup/gm03:GM03_2_1Core.Core.PT_URLGroup', namespaces)):
            pt_groups.append(PT_Group(pt_group))

        self.pt_group = pt_groups


class GM03(object):
    """TRANSFER parser"""
    def __init__(self, md):
        """constructor"""

        if hasattr(md, 'getroot'):  # standalone document
            self.xml = etree.tostring(md.getroot())
        else:  # part of a larger document
            self.xml = etree.tostring(md)

        self.header = HeaderSection(md)
        self.data = DataSection(md)


class HeaderSection(object):
    """HEADERSECTION parser"""
    def __init__(self, md):
        """constructor"""

        header = None
        header = md.find(util.nspath_eval('gm03:HEADERSECTION', namespaces))

        if header is None:
            return None

        self.version = header.attrib.get('VERSION')
        self.sender = header.attrib.get('SENDER')
        self.models = []

        for model in header.findall(util.nspath_eval('gm03:MODELS/gm03:MODEL', namespaces)):
            name = util.testXMLValue(model.find(util.nspath_eval('gm03:NAME', namespaces)))
            version = util.testXMLValue(model.find(util.nspath_eval('gm03:VERSION', namespaces)))
            uri = util.testXMLValue(model.find(util.nspath_eval('gm03:URI', namespaces)))
            model_dict = {
                'name': name,
                'version': version,
                'uri': uri
            }
            self.models.append(model_dict)

        self.comment = util.testXMLValue(header.find(util.nspath_eval('gm03:COMMENT', namespaces)))


class DataSection(object):
    """DATASECTION parser"""
    def __init__(self, md):
        """constructor"""

        section = None
        section = md.find(util.nspath_eval('gm03:DATASECTION', namespaces))

        if section is None:
            return None

        mdata = section.find(util.nspath_eval('gm03:GM03_2_1Core.Core', namespaces))
        if mdata is not None:
            self.core = Core(mdata)
        else:
            mdata = section.find(util.nspath_eval('gm03:GM03_2_1Comprehensive.Comprehensive', namespaces))
            if mdata is not None:
                self.comprehensive = Comprehensive(mdata)


class Core(object):
    """Core parser"""
    def __init__(self, md):
        """constructor"""

        self.bid = md.attrib.get('BID')

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.DQ_DataQuality', namespaces))
        if val is not None:
            self.data_quality = DQ_DataQuality(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Authority', namespaces))
        if val is not None:
            self.authority = _GenericObject(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_DigitalTransferOptions', namespaces))
        if val is not None:
            self.digital_transfer_options = MD_DigitalTransferOptions(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Distribution', namespaces))
        if val is not None:
            self.distribution = _GenericObject(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_ReferenceSystem', namespaces))
        if val is not None:
            self.reference_system_identifier = MD_ReferenceSystem(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Thesaurus', namespaces))
        if val is not None:
            self.thesaurus = MD_Thesaurus(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.SC_VerticalDatum', namespaces))
        if val is not None:
            self.vertical_datum = SC_VerticalDatum(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_Address', namespaces))
        if val is not None:
            self.address = CI_Address(val)

        self.date = []
        for cid in md.findall(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_Date', namespaces)):
            self.date.append(CI_Date(cid))

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_Telephone', namespaces))
        if val is not None:
            self.telephone = CI_Telephone(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.DQ_Scope', namespaces))
        if val is not None:
            self.scope = DQ_Scope(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_VerticalExtent', namespaces))
        if val is not None:
            self.vertical_extent = EX_VerticalExtent(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Format', namespaces))
        if val is not None:
            self.format = MD_Format(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Metadata', namespaces))
        if val is not None:
            self.metadata = MD_Metadata(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_RepresentativeFraction', namespaces))
        if val is not None:
            self.representative_fraction = MD_RepresentativeFraction(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Resolution', namespaces))
        if val is not None:
            self.resolution = MD_Resolution(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_BoundingPolygon', namespaces))
        if val is not None:
            self.bounding_polygon = EX_BoundingPolygon(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_GeographicBoundingBox', namespaces))
        if val is not None:
            self.geographic_bounding_box = EX_GeographicBoundingBox(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_GeographicDescription', namespaces))
        if val is not None:
            self.geographic_description = EX_GeographicDescription(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_TemporalExtent', namespaces))
        if val is not None:
            self.temporal_extent = EX_TemporalExtent(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_DistributiondistributionFormat', namespaces))
        if val is not None:
            self.distribution_distribution_format = MD_DistributiondistributionFormat(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.referenceSystemInfoMD_Metadata', namespaces))
        if val is not None:
            self.reference_system_metadata = referenceSystemInfoMD_Metadata(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_Citation', namespaces))
        if val is None:
            val = md.find(util.nspath_eval('gm03:GM03_2_1Comprehensive.Comprehensive.CI_Citation', namespaces))
            if val is not None:
                self.citation = CI_Citation(val)
        else:
            self.citation = CI_Citation(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_Contact', namespaces))
        if val is not None:
            self.contact = CI_Contact(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_OnlineResource', namespaces))
        if val is not None:
            self.online_resource = CI_OnlineResource(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_ResponsibleParty', namespaces))
        if val is not None:
            self.responsible_party = CI_ResponsibleParty(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_Extent', namespaces))
        if val is not None:
            self.extent = EX_Extent(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_SpatialTemporalExtent', namespaces))
        if val is not None:
            self.spatial_temporal_extent = EX_Extent(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.LI_Lineage', namespaces))
        if val is not None:
            self.lineage = LI_Lineage(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Identifier', namespaces))
        if val is not None:
            self.identifier = MD_Identifier(val)

        self.keywords = []
        for kw in md.findall(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Keywords', namespaces)):
            self.keywords.append(MD_Keywords(kw))

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_DataIdentification', namespaces))
        if val is None:
            val = md.find(util.nspath_eval(
                'gm03:GM03_2_1Comprehensive.Comprehensive.MD_DataIdentification', namespaces))
            if val is not None:
                self.data_identification = MD_DataIdentification(val)
        else:
            self.data_identification = MD_DataIdentification(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.RS_Identifier', namespaces))
        if val is not None:
            self.rs_identifier = RS_Identifier(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.CI_ResponsiblePartyparentinfo', namespaces))
        if val is not None:
            self.responsible_party_parent_info = CI_ResponsiblePartyparentinfo(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.descriptiveKeywordsMD_Identification', namespaces))
        if val is not None:
            self.descriptive_keywords_identification = descriptiveKeywordsMD_Identification(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_ExtentgeographicElement', namespaces))
        if val is not None:
            self.extent_geographic_element = EX_ExtentgeographicElement(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_ExtenttemporalElement', namespaces))
        if val is not None:
            self.extent_temporal_element = EX_ExtenttemporalElement(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.EX_ExtentverticalElement', namespaces))
        if val is not None:
            self.extent_vertical_element = EX_ExtenttemporalElement(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_IdentificationpointOfContact', namespaces))
        if val is not None:
            self.identification_point_of_contact = MD_IdentificationpointOfContact(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_Metadatacontact', namespaces))
        if val is not None:
            self.metadata_point_of_contact = MD_Metadatacontact(val)

        val = md.find(util.nspath_eval('gm03:GM03_2_1Core.Core.spatialExtentEX_SpatialTemporalExtent', namespaces))
        if val is not None:
            self.self.spatial_temporal_extent = spatialExtentEX_SpatialTemporalExtent(val)

    @property
    def elements(self):
        """helper function to return all properties"""

        dict_elements = vars(self)
        if 'bid' in dict_elements:
            del dict_elements['bid']
        return dict_elements

    def get_element_by_tid(self, tid):
        """helper function to find values by reference"""

        for key, value in list(self.elements.items()):
            if hasattr(value, 'tid') and value.tid == tid:
                return self.elements[key]
        return None


class DQ_DataQuality(_GenericObject):
    """DQ_DataQuality parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:MD_Metadata', namespaces))
        self.metadata = _GenericObjectProperty(val)


class MD_Authority(_GenericObject):
    """MD_Authority parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)


class MD_ReferenceSystem(_GenericObject):
    """MD_ReferenceSystem parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:referenceSystemIdentifier', namespaces))
        self.reference_system_identifier = _GenericObjectProperty(val)


class MD_DigitalTransferOptions(_GenericObject):
    """MD_DigitalTransferOptions parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:MD_Distribution', namespaces))
        self.distribution = _GenericObjectProperty(val)


class MD_Thesaurus(_GenericObject):
    """MD_Thesaurus parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:citation', namespaces))
        self.citation = _GenericObjectProperty(val)


class SC_VerticalDatum(_GenericObject):
    """SC_VerticalDatum parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:datumID', namespaces))
        self.datum_id = _GenericObjectProperty(val)


class CI_Address(_GenericObject):
    """CI_Address parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.street_name = util.testXMLValue(md.find(util.nspath_eval('gm03:streetName', namespaces)))
        self.street_number = util.testXMLValue(md.find(util.nspath_eval('gm03:streetNumber', namespaces)))
        self.address_line = util.testXMLValue(md.find(util.nspath_eval('gm03:addressLine', namespaces)))
        self.post_box = util.testXMLValue(md.find(util.nspath_eval('gm03:postBox', namespaces)))
        self.postal_code = util.testXMLValue(md.find(util.nspath_eval('gm03:postalCode', namespaces)))
        self.city = util.testXMLValue(md.find(util.nspath_eval('gm03:city', namespaces)))
        self.administrative_area = util.testXMLValue(md.find(util.nspath_eval('gm03:administrativeArea', namespaces)))
        self.country = util.testXMLValue(md.find(util.nspath_eval('gm03:country', namespaces)))


class CI_Date(_GenericObject):
    """CI_Date parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.date = util.testXMLValue(md.find(util.nspath_eval('gm03:date', namespaces)))
        self.date_type = util.testXMLValue(md.find(util.nspath_eval('gm03:dateType', namespaces)))

        val = md.find(util.nspath_eval('gm03:CI_Citation', namespaces))
        self.citation = _GenericObjectProperty(val)


class CI_Telephone(_GenericObject):
    """CI_Telephone parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.number = util.testXMLValue(md.find(util.nspath_eval('gm03:number', namespaces)))
        self.number_type = util.testXMLValue(md.find(util.nspath_eval('gm03:numberType', namespaces)))

        val = md.find(util.nspath_eval('gm03:CI_ResponsibleParty', namespaces))
        self.responsible_party = _GenericObjectProperty(val)


class DQ_Scope(_GenericObject):
    """DQ_Scope parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.level = util.testXMLValue(md.find(util.nspath_eval('gm03:level', namespaces)))

        val = md.find(util.nspath_eval('gm03:DQ_DataQuality', namespaces))
        self.data_quality = _GenericObjectProperty(val)


class EX_VerticalExtent(_GenericObject):
    """EX_VerticalExtent parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.minimum_value = util.testXMLValue(md.find(util.nspath_eval('gm03:minimumValue', namespaces)))
        self.maximum_value = util.testXMLValue(md.find(util.nspath_eval('gm03:maximumValue', namespaces)))
        self.unit_of_measure = util.testXMLValue(md.find(util.nspath_eval('gm03:unitOfMeasure', namespaces)))

        val = md.find(util.nspath_eval('gm03:verticalDatum', namespaces))
        self.vertical_datum = _GenericObjectProperty(val)


class MD_Format(_GenericObject):
    """MD_Format parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.name = util.testXMLValue(md.find(util.nspath_eval('gm03:name', namespaces)))
        self.version = util.testXMLValue(md.find(util.nspath_eval('gm03:version', namespaces)))


class MD_Metadata(_GenericObject):
    """MD_Metadata parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.file_identifier = util.testXMLValue(md.find(util.nspath_eval('gm03:fileIdentifier', namespaces)))
        self.language = util.testXMLValue(md.find(util.nspath_eval('gm03:language', namespaces)))
        self.character_set = util.testXMLValue(md.find(util.nspath_eval('gm03:characterSet', namespaces)))
        self.date_stamp = util.testXMLValue(md.find(util.nspath_eval('gm03:dateStamp', namespaces)))
        self.metadata_standard_name = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:metadataStandardName', namespaces)))
        self.metadata_standard_version = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:metadataStandardVersion', namespaces)))
        self.dataset_uri = util.testXMLValue(md.find(util.nspath_eval('gm03:dataSetURI', namespaces)))

        val = md.find(util.nspath_eval('gm03:hierarchyLevel', namespaces))
        if val is not None:
            values = []
            for value in val.findall(util.nspath_eval('gm03:GM03_2_1Core.Core.MD_ScopeCode_/gm03:value', namespaces)):
                values.append(util.testXMLValue(value))
            self.hierarchy_level = values

        val = md.find(util.nspath_eval('gm03:hierarchyLevelName', namespaces))
        if val is not None:
            values = []
            for value in val.findall(util.nspath_eval(
                    'gm03:GM03_2_1Core.Core.CharacterString_/gm03:value', namespaces)):
                values.append(util.testXMLValue(value))
            self.hierarchy_level_name = values

        val = md.find(util.nspath_eval('gm03:distributionInfo', namespaces))
        self.distribution_info = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:parentIdentifier', namespaces))
        self.parent_identifier = _GenericObjectProperty(val)


class MD_RepresentativeFraction(_GenericObject):
    """MD_RepresentativeFraction parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.denominator = util.testXMLValue(md.find(util.nspath_eval('gm03:denominator', namespaces)))


class MD_Resolution(_GenericObject):
    """MD_Resolution parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.distance = util.testXMLValue(md.find(util.nspath_eval('gm03:distance', namespaces)))

        val = md.find(util.nspath_eval('gm03:MD_DataIdentification', namespaces))
        self.data_identification = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:equivalentScale', namespaces))
        self.equivalent_scale = _GenericObjectProperty(val)


class MD_ScopeDescription(_GenericObject):
    """MD_ScopeDescription parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.attributes = util.testXMLValue(md.find(util.nspath_eval('gm03:attributes', namespaces)))
        self.features = util.testXMLValue(md.find(util.nspath_eval('gm03:features', namespaces)))
        self.feature_instances = util.testXMLValue(md.find(util.nspath_eval('gm03:featureInstances', namespaces)))
        self.attribute_instances = util.testXMLValue(md.find(util.nspath_eval('gm03:attributeInstances', namespaces)))
        self.dataset = util.testXMLValue(md.find(util.nspath_eval('gm03:dataset', namespaces)))
        self.other = util.testXMLValue(md.find(util.nspath_eval('gm03:other', namespaces)))

        val = md.find(util.nspath_eval('gm03:DQ_Scope', namespaces))
        self.scope = _GenericObjectProperty(val)


class EX_BoundingPolygon(_GenericObject):
    """EX_BoundingPolygon parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.extent_type_code = util.testXMLValue(md.find(util.nspath_eval('gm03:extentTypeCode', namespaces)))

        boundaries = []

        for boundary in md.findall(util.nspath_eval('gm03:SURFACE/gm03:BOUNDARY', namespaces)):
            polylines = []
            for polyline in boundary.findall(util.nspath_eval('gm03:POLYLINE', namespaces)):
                coords = []
                arcs = []
                for coord in polyline.findall(util.nspath_eval('gm03:COORD', namespaces)):
                    c1 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C1', namespaces)))
                    c2 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C2', namespaces)))
                    c3 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C3', namespaces)))
                    coordvalue = {'c1': c1, 'c2': c2, 'c3': c3}
                    coords.append(coordvalue)
                for arc in polyline.findall(util.nspath_eval('gm03:ARC', namespaces)):
                    c1 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C1', namespaces)))
                    c2 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C2', namespaces)))
                    c3 = util.testXMLValue(coord.find(util.nspath_eval('gm03:C3', namespaces)))
                    a1 = util.testXMLValue(coord.find(util.nspath_eval('gm03:A1', namespaces)))
                    a2 = util.testXMLValue(coord.find(util.nspath_eval('gm03:A2', namespaces)))
                    r = util.testXMLValue(coord.find(util.nspath_eval('gm03:R', namespaces)))
                    arcpoint = {'c1': c1, 'c2': c2, 'c3': c3, 'a1': a1, 'a2': a2, 'r': r}
                    arcs.append(arcpoint)
                polylines.append(coords)
                polylines.append(arcs)
            boundaries.append(polylines)
        self.boundary = boundaries


class EX_GeographicBoundingBox(_GenericObject):
    """EX_GeographicBoundingBox parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.extent_type_code = util.testXMLValue(md.find(util.nspath_eval('gm03:extentTypeCode', namespaces)))
        self.north_bound_latitude = util.testXMLValue(md.find(util.nspath_eval('gm03:northBoundLatitude', namespaces)))
        self.south_bound_latitude = util.testXMLValue(md.find(util.nspath_eval('gm03:southBoundLatitude', namespaces)))
        self.east_bound_longitude = util.testXMLValue(md.find(util.nspath_eval('gm03:eastBoundLongitude', namespaces)))
        self.west_bound_longitude = util.testXMLValue(md.find(util.nspath_eval('gm03:westBoundLongitude', namespaces)))


class EX_GeographicDescription(_GenericObject):
    """EX_GeographicDescription parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.extent_type_code = util.testXMLValue(md.find(util.nspath_eval('gm03:extentTypeCode', namespaces)))

        val = md.find(util.nspath_eval('gm03:geographicIdentifier', namespaces))
        self.geographic_identifier = _GenericObjectProperty(val)


class EX_TemporalExtent(_GenericObject):
    """EX_TemporalExtent parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        begin = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:extent/gm03:GM03_2_1Core.Core.TM_Primitive/begin', namespaces)))
        end = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:extent/gm03:GM03_2_1Core.Core.TM_Primitive/end', namespaces)))

        self.extent = {'begin': begin, 'end': end}


class MD_DistributiondistributionFormat(object):
    """MD_DistributiondistributionFormat parser"""
    def __init__(self, md):
        """constructor"""

        val = md.find(util.nspath_eval('gm03:MD_Distribution', namespaces))
        self.distribution = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:distributionFormat', namespaces))
        self.distribution_format = _GenericObjectProperty(val)


class referenceSystemInfoMD_Metadata(object):
    """referenceSystemInfoMD_Metadata parser"""
    def __init__(self, md):
        """constructor"""

        val = md.find(util.nspath_eval('gm03:referenceSystemInfo', namespaces))
        self.reference_system_info = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:MD_Metadata', namespaces))
        self.metadata = _GenericObjectProperty(val)


class CI_Citation(_GenericObject):
    """CI_Citation parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:title', namespaces))
        if val is not None:
            self.title = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:MD_Authority', namespaces))
        self.authority = _GenericObjectProperty(val)


class CI_Contact(_GenericObject):
    """CI_Contact parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.hours_of_service = util.testXMLValue(md.find(util.nspath_eval('gm03:hoursOfService', namespaces)))

        val = md.find(util.nspath_eval('gm03:contactInstructions', namespaces))
        if val is not None:
            self.contact_instructions = PT_FreeText(val)


class CI_OnlineResource(_GenericObject):
    """CI_OnlineResource parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.protocol = util.testXMLValue(md.find(util.nspath_eval('gm03:protocol', namespaces)))
        self.application_profile = util.testXMLValue(md.find(util.nspath_eval('gm03:applicationProfile', namespaces)))
        self.function = util.testXMLValue(md.find(util.nspath_eval('gm03:function', namespaces)))

        val = md.find(util.nspath_eval('gm03:description', namespaces))
        if val is not None:
            self.description = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:name', namespaces))
        if val is not None:
            self.name = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:linkage', namespaces))
        if val is not None:
            self.linkage = PT_FreeURL(val)

        val = md.find(util.nspath_eval('gm03:MD_DigitalTransferOptions', namespaces))
        self.digital_transfer_options = _GenericObjectProperty(val)


class CI_ResponsibleParty(_GenericObject):
    """CI_ResponsibleParty parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.individual_first_name = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:individualFirstName', namespaces)))
        self.individual_last_name = util.testXMLValue(md.find(util.nspath_eval(
            'gm03:individualLastName', namespaces)))

        val = md.find(util.nspath_eval('gm03:electronicalMailAddress', namespaces))
        if val is not None:
            self.electronical_mail_address = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.URL_/gm03:value', namespaces)))

        val = md.find(util.nspath_eval('gm03:organisationName', namespaces))
        if val is not None:
            self.organisation_name = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:positionName', namespaces))
        if val is not None:
            self.position_name = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:organisationAcronym', namespaces))
        if val is not None:
            self.organisation_acronym = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:linkage', namespaces))
        if val is not None:
            self.linkage = PT_FreeURL(val)

        val = md.find(util.nspath_eval('gm03:address', namespaces))
        self.address = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:contactInfo', namespaces))
        self.contact_info = _GenericObjectProperty(val)


class EX_Extent(_GenericObject):
    """EX_Extent parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:description', namespaces))
        if val is not None:
            self.description = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:MD_DataIdentification', namespaces))
        self.data_identification = _GenericObjectProperty(val)


class EX_SpatialTemporalExtent(EX_TemporalExtent):
    """EX_SpatialTemporalExtent parser"""
    def __init__(self, md):
        """constructor"""

        EX_TemporalExtent.__init__(self, md)


class LI_Lineage(_GenericObject):
    """LI_Lineage parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:statement', namespaces))
        if val is not None:
            self.statement = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:DQ_DataQuality', namespaces))
        self.data_quality = _GenericObjectProperty(val)


class MD_Identifier(_GenericObject):
    """MD_Identifier parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:code', namespaces))
        if val is not None:
            self.code = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:MD_Authority', namespaces))
        self.authority = _GenericObjectProperty(val)


class MD_Keywords(_GenericObject):
    """MD_Keywords parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        self.type = util.testXMLValue(md.find(util.nspath_eval('gm03:type', namespaces)))

        val = md.find(util.nspath_eval('gm03:keyword', namespaces))
        if val is not None:
            self.keyword = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:thesaurus', namespaces))
        self.thesaurus = _GenericObjectProperty(val)


class MD_DataIdentification(_GenericObject):
    """MD_DataIdentification parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:status', namespaces))
        if val is not None:
            self.status = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.MD_ProgressCode_/gm03:value', namespaces)))

        val = md.find(util.nspath_eval('gm03:abstract', namespaces))
        if val is not None:
            self.abstract = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:purpose', namespaces))
        if val is not None:
            self.purpose = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:MD_Metadata', namespaces))
        self.metadata = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:citation', namespaces))
        self.citation = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:spatialRepresentationType', namespaces))
        if val is not None:
            self.spatial_representation_type = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.MD_SpatialRepresentationTypeCode_/gm03:value', namespaces)))

        val = md.find(util.nspath_eval('gm03:language', namespaces))
        if val is not None:
            self.language = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:CodeISO.LanguageCodeISO_/gm03:value', namespaces)))

        val = md.find(util.nspath_eval('gm03:characterSet', namespaces))
        if val is not None:
            self.character_set = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.MD_CharacterSetCode_/gm03:value', namespaces)))

        val = md.find(util.nspath_eval('gm03:topicCategory', namespaces))
        if val is not None:
            self.topic_category = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.MD_TopicCategoryCode_/gm03:value', namespaces)))


class RS_Identifier(_GenericObject):
    """RS_Identifier parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:code', namespaces))
        if val is not None:
            self.code = PT_FreeText(val)

        val = md.find(util.nspath_eval('gm03:MD_Authority', namespaces))
        self.authority = _GenericObjectProperty(val)


class CI_ResponsiblePartyparentinfo(_GenericObject):
    """CI_ResponsiblePartyparentinfo parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:parentResponsibleParty', namespaces))
        self.parent_responsible_party = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:CI_ResponsibleParty', namespaces))
        self.responsible_party = _GenericObjectProperty(val)


class descriptiveKeywordsMD_Identification(_GenericObject):
    """descriptiveKeywordsMD_Identification parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:descriptiveKeywords', namespaces))
        self.descriptive_keywords = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:MD_Identification', namespaces))
        self.identification = _GenericObjectProperty(val)


class EX_ExtenttemporalElement(_GenericObject):
    """EX_ExtenttemporalElement parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:EX_Extent', namespaces))
        self.extent = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:temporalElement', namespaces))
        self.temporal_element = _GenericObjectProperty(val)


class EX_ExtentverticalElement(_GenericObject):
    """EX_ExtentverticalElement parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:EX_Extent', namespaces))
        self.extent = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:verticalElement', namespaces))
        self.vertical_element = _GenericObjectProperty(val)


class EX_ExtentgeographicElement(_GenericObject):
    """EX_ExtentgeographicElement parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:EX_Extent', namespaces))
        self.extent = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:geographicElement', namespaces))
        self.geographic_element = _GenericObjectProperty(val)


class MD_IdentificationpointOfContact(_GenericObject):
    """MD_IdentificationpointOfContact parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:pointOfContact', namespaces))
        self.point_of_contact = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:MD_Identification', namespaces))
        self.identification = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:role', namespaces))
        if val is not None:
            self.role = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.CI_RoleCode_/gm03:value', namespaces)))


class MD_Metadatacontact(_GenericObject):
    """MD_Metadatacontact parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:contact', namespaces))
        self.contact = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:MD_Metadata', namespaces))
        self.metadata = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:role', namespaces))
        if val is not None:
            self.role = util.testXMLValue(val.find(util.nspath_eval(
                'gm03:GM03_2_1Core.Core.CI_RoleCode_/gm03:value', namespaces)))


class spatialExtentEX_SpatialTemporalExtent(_GenericObject):
    """spatialExtentEX_SpatialTemporalExtent parser"""
    def __init__(self, md):
        """constructor"""

        _GenericObject.__init__(self, md)

        val = md.find(util.nspath_eval('gm03:spatialExtent', namespaces))
        self.spatial_extent = _GenericObjectProperty(val)

        val = md.find(util.nspath_eval('gm03:EX_SpatialTemporalExtent', namespaces))
        self.spatial_temporal_extent = _GenericObjectProperty(val)


class Comprehensive(Core):
    """Comprehensive parser"""
    def __init__(self, md):
        """constructor"""

        Core.__init__(self, md)



from dataclasses import dataclass
from typing import Sequence
from owslib.etree import etree
from owslib import util
from owslib.namespaces import Namespaces


def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces(["gml32", "ogc", "xsd"])
    ns[None] = n.get_namespace("ogc")
    return ns


namespaces = get_namespaces()


def prefix(x):
    """Shorthand to insert namespaces."""
    return util.nspath_eval(x, namespaces)


@dataclass
class AbstractGMLType():
    id: str


@dataclass
class AbstractGeometryType(AbstractGMLType):
    srsName: str = None
    srsDimension: int = None
    # axisLabels: str = None
    # uomLabels: str = None

    description: str = None
    descriptionReference: str = None
    identifier: str = None
    name: str = None


@dataclass
class _PointBase:
    """This is used to avoid issues arising from non-optional attributes defined after optional attributes."""
    pos: Sequence[float]


@dataclass
class Point(AbstractGeometryType, _PointBase):
    """GML Point object."""

    def toXML(self):
        """Return `lxml.etree.Element` object."""
        node = etree.Element(prefix("gml32:Point"))
        for key in ["id", "srsName"]:
            if getattr(self, key, None) is not None:
                node.set(prefix(f"gml32:{key}"), getattr(self, key))

        for key in ["description", "descriptionReference", "identifier", "name"]:
            content = getattr(self, key)
            if content is not None:
                etree.SubElement(node, prefix(f"gml32:{key}")).text = content

        coords = etree.SubElement(node, prefix("gml32:pos"))
        coords.text = " ".join([str(c) for c in self.pos])
        for key in ["srsDimension"]:
            if getattr(self, key, None) is not None:
                node.set(prefix(f"gml32:{key}"), getattr(self, key))

        return node



# =============================================================================
# Copyright (c) 2023 Tom Kralidis
#
# Author: Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

from io import BytesIO
import logging
from typing import BinaryIO

from owslib.ogcapi.features import Features
from owslib.util import Authentication

LOGGER = logging.getLogger(__name__)


class EnvironmentalDataRetrieval(Features):
    """Abstraction for OGC API - Environmental Data Retrieval"""

    def __init__(self, url: str, json_: str = None, timeout: int = 30,
                 headers: dict = None, auth: Authentication = None):
        __doc__ = Features.__doc__  # noqa
        super().__init__(url, json_, timeout, headers, auth)

    def data(self) -> list:
        """
        implements /collections filtered on EDR data resources

        @returns: `list` of filtered collections object
        """

        datas = []
        collections_ = super().collections()

        for c_ in collections_['collections']:
            for l_ in c_['links']:
                if 'data' in l_['rel']:
                    datas.append(c_['id'])
                    break

        return datas

    def query_data(self, collection_id: str,
                   query_type: str, **kwargs: dict) -> BinaryIO:
        """
        implements /collection/{collectionId}/coverage/

        @type collection_id: string
        @param collection_id: id of collection
        @type query_type: string
        @param query_type: query type
        @type bbox: list
        @param bbox: list of minx,miny,maxx,maxy
        @type coords: string
        @param coords: well-known text geometry
        @type datetime_: string
        @type datetime_: string
        @param datetime_: time extent or time instant
        @type parameter_names: list
        @param parameter_names: list of parameter names

        @returns: coverage data
        """

        kwargs_ = {}

        if 'bbox' in kwargs:
            kwargs_['bbox'] = ','.join(list(map(str, kwargs['bbox'])))
        if 'parameter_names' in kwargs:
            kwargs_['parameter_names'] = ','.join(kwargs['parameter_names'])

        query_args_map = {
            'coords': 'coords',
            'corridor_width': 'corridor-width',
            'corridor_height': 'corridor-height',
            'crs': 'crs',
            'cube-z': 'cube-z',
            'datetime_': 'datetime',
            'height': 'height',
            'height_units': 'height-units',
            'resolution_x': 'resolution-x',
            'resolution_y': 'resolution-y',
            'resolution_z': 'resolution-z',
            'width': 'width',
            'width_units': 'width-units',
            'within': 'within',
            'z': 'z'
        }

        for key, value in query_args_map.items():
            if key in kwargs:
                kwargs_[value] = kwargs[key]

        path = f'collections/{collection_id}/{query_type}'

        return self._request(path=path, kwargs=kwargs_)



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2009 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

""" DIF metadata parser """

from owslib.etree import etree
from owslib import util
from owslib.namespaces import Namespaces


# default variables
def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces("dif")
    ns[None] = n.get_namespace("dif")
    return ns


namespaces = get_namespaces()


class DIF(object):
    """ Process DIF """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Entry_ID', namespaces))
        self.identifier = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Entry_Title', namespaces))
        self.title = util.testXMLValue(val)

        self.citation = []
        for el in md.findall(util.nspath_eval('dif:Data_Set_Citation', namespaces)):
            self.citation.append(Citation(el))

        self.personnel = []
        for el in md.findall(util.nspath_eval('dif:Personnel', namespaces)):
            self.personnel.append(util.testXMLValue(el))

        self.discipline = []
        for el in md.findall(util.nspath_eval('dif:Discipline', namespaces)):
            self.discipline.append(util.testXMLValue(el))

        self.parameters = []
        for el in md.findall(util.nspath_eval('dif:Parameters', namespaces)):
            self.parameters.append(util.testXMLValue(el))

        self.iso_topic_category = []
        for el in md.findall(util.nspath_eval('dif:ISO_Topic_Category', namespaces)):
            self.iso_topic_category.append(util.testXMLValue(el))

        self.keyword = []
        for el in md.findall(util.nspath_eval('dif:Keyword', namespaces)):
            self.keyword.append(util.testXMLValue(el))

        self.sensor_name = []
        for el in md.findall(util.nspath_eval('dif:Sensor_Name', namespaces)):
            self.sensor_name.append(Name(el))

        self.source_name = []
        for el in md.findall(util.nspath_eval('dif:Source_Name', namespaces)):
            self.source_name.append(Name(el))

        self.temporal_coverage = []
        for el in md.findall(util.nspath_eval('dif:Temporal_Coverage', namespaces)):
            self.temporal_coverage.append(Temporal_Coverage(el))

        self.paleo_temporal_coverage = []
        for el in md.findall(util.nspath_eval('dif:Paleo_Temporal_Coverage', namespaces)):
            self.paleo_temporal_coverage.append(Paleo_Temporal_Coverage(el))

        self.data_set_progress = []
        for el in md.findall(util.nspath_eval('dif:Data_Set_Progress', namespaces)):
            self.data_set_progress.append(util.testXMLValue(el))

        self.spatial_coverage = []
        for el in md.findall(util.nspath_eval('dif:Spatial_Coverage', namespaces)):
            self.spatial_coverage.append(Spatial_Coverage(el))

        self.location = []
        for el in md.findall(util.nspath_eval('dif:location', namespaces)):
            self.location.append(util.testXMLValue(el))

        self.data_resolution = []
        for el in md.findall(util.nspath_eval('dif:Data_Resolution', namespaces)):
            self.data_resolution.append(Data_Resolution(el))

        self.project = []
        for el in md.findall(util.nspath_eval('dif:Project', namespaces)):
            self.project.append(Name(el))

        val = md.find(util.nspath_eval('dif:Quality', namespaces))
        self.quality = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Access_Constraints', namespaces))
        self.access_constraints = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Use_Constraints', namespaces))
        self.use_constraints = util.testXMLValue(val)

        self.language = []
        for el in md.findall(util.nspath_eval('dif:Data_Set_Language', namespaces)):
            self.language.append(util.testXMLValue(el))

        self.originating_center = []
        for el in md.findall(util.nspath_eval('dif:Originating_Center', namespaces)):
            self.originating_center.append(util.testXMLValue(el))

        self.data_center = []
        for el in md.findall(util.nspath_eval('dif:Data_Center', namespaces)):
            self.data_center.append(Data_Center(el))

        self.distribution = []
        for el in md.findall(util.nspath_eval('dif:Distribution', namespaces)):
            self.distribution.append(Distribution(el))

        self.multimedia_sample = []
        for el in md.findall(util.nspath_eval('dif:Multimedia_Sample', namespaces)):
            self.multimedia_sample.append(Multimedia_Sample(el))

        val = md.find(util.nspath_eval('dif:Reference', namespaces))
        self.reference = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Summary', namespaces))
        self.summary = util.testXMLValue(val)

        self.related_url = []
        for el in md.findall(util.nspath_eval('dif:Related_URL', namespaces)):
            self.related_url.append(Related_URL(el))

        self.parent_dif = []
        for el in md.findall(util.nspath_eval('dif:Parent_DIF', namespaces)):
            self.parent_dif.append(util.testXMLValue(el))

        self.idn_node = []
        for el in md.findall(util.nspath_eval('dif:IDN_Node', namespaces)):
            self.idn_node.append(Name(el))

        val = md.find(util.nspath_eval('dif:Originating_Metadata_Node', namespaces))
        self.originating_metadata_node = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Metadata_Name', namespaces))
        self.metadata_name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Metadata_Version', namespaces))
        self.metadata_version = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:DIF_Creation_Date', namespaces))
        self.dif_creation_date = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Last_DIF_Revision_Date', namespaces))
        self.last_dif_revision_date = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Future_DIF_Review_Date', namespaces))
        self.future_dif_review_date = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Private', namespaces))
        self.private = util.testXMLValue(val)


class Citation(object):
    """ Parse Data_Set_Citation """
    def __init__(self, el):
        val = el.find(util.nspath_eval('dif:Dataset_Creator', namespaces))
        self.creator = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Dataset_Title', namespaces))
        self.title = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Dataset_Series_Name', namespaces))
        self.series_name = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Dataset_Release_Date', namespaces))
        self.release_date = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Dataset_Release_Place', namespaces))
        self.release_place = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Dataset_Publisher', namespaces))
        self.publisher = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Version', namespaces))
        self.version = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Issue_Identification', namespaces))
        self.issue_identification = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Data_Presentation_Form', namespaces))
        self.presentation_form = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Other_Citation_Details', namespaces))
        self.details = util.testXMLValue(val)

        val = el.find(util.nspath_eval('dif:Online_Resource', namespaces))
        self.onlineresource = util.testXMLValue(val)


class Personnel(object):
    """ Process Personnel """
    def __init__(self, md):
        self.role = []
        for el in md.findall(util.nspath_eval('dif:Role', namespaces)):
            self.role.append(util.testXMLValue(el))

        val = md.find(util.nspath_eval('dif:First_Name', namespaces))
        self.first_name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Middle_Name', namespaces))
        self.middle_name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Last_Name', namespaces))
        self.last_name = util.testXMLValue(val)

        self.email = []
        for el in md.findall(util.nspath_eval('dif:Email', namespaces)):
            self.email.append(util.testXMLValue(el))

        self.phone = []
        for el in md.findall(util.nspath_eval('dif:Phone', namespaces)):
            self.phone.append(util.testXMLValue(el))

        self.fax = []
        for el in md.findall(util.nspath_eval('dif:Fax', namespaces)):
            self.fax.append(util.testXMLValue(el))

        val = md.find(util.nspath_eval('dif:Contact_Address', namespaces))
        self.contact_address = Contact_Address(val)


class Contact_Address(object):
    """ Process Contact_Address """
    def __init__(self, md):
        self.address = []
        for el in md.findall(util.nspath_eval('dif:Address', namespaces)):
            self.address.append(util.testXMLValue(el))

        val = md.find(util.nspath_eval('dif:City', namespaces))
        self.city = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Province_or_State', namespaces))
        self.province_or_state = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Postal_Code', namespaces))
        self.postal_code = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Country', namespaces))
        self.country = util.testXMLValue(val)


class Discipline(object):
    """ Process Discipline """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Discipline_Name', namespaces))
        self.name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Subdiscipline', namespaces))
        self.subdiscipline = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Detailed_Subdiscipline', namespaces))
        self.detailed_subdiscipline = util.testXMLValue(val)


class Parameters(object):
    """ Process Parameters """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Category', namespaces))
        self.category = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Topic', namespaces))
        self.topic = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Term', namespaces))
        self.term = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Variable_Level_1', namespaces))
        self.variable_l1 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Variable_Level_2', namespaces))
        self.variable_l2 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Variable_Level_3', namespaces))
        self.variable_l3 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Detailed_Variable', namespaces))
        self.detailed_variable = util.testXMLValue(val)


class Name(object):
    """ Process Sensor_Name, Source_Name, Project, IDN_Node """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Short_Name', namespaces))
        self.short_name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Long_Name', namespaces))
        self.long_name = util.testXMLValue(val)


class Temporal_Coverage(object):
    """ Process Temporal_Coverage """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Start_Date', namespaces))
        self.start_date = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:End_Date', namespaces))
        self.end_date = util.testXMLValue(val)


class Paleo_Temporal_Coverage(object):
    """ Process Paleo_Temporal_Coverage """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Paleo_Start_Date', namespaces))
        self.paleo_start_date = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Paleo_End_Date', namespaces))
        self.paleo_end_date = util.testXMLValue(val)

        self.chronostratigraphic_unit = []
        for el in md.findall(util.nspath_eval('dif:Chronostratigraphic_Unit', namespaces)):
            self.chronostratigraphic_unit.append(Chronostratigraphic_Unit(el))


class Chronostratigraphic_Unit(object):
    """ Process Chronostratigraphic_Unit """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Eon', namespaces))
        self.eon = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Era', namespaces))
        self.era = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Period', namespaces))
        self.period = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Epoch', namespaces))
        self.epoch = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Stage', namespaces))
        self.stage = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Detailed_Classification', namespaces))
        self.detailed_classification = util.testXMLValue(val)


class Spatial_Coverage(object):
    """ Process Spatial_Coverage """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Southernmost_Latitude', namespaces))
        self.miny = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Northernmost_Latitude', namespaces))
        self.maxy = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Westernmost_Latitude', namespaces))
        self.minx = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Easternmost_Latitude', namespaces))
        self.maxx = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Minimum_Altitude', namespaces))
        self.minz = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Maximum_Altitude', namespaces))
        self.maxz = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Minimum_Depth', namespaces))
        self.mindepth = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Maximum_Depth', namespaces))
        self.maxdepth = util.testXMLValue(val)


class Location(object):
    """ Process Location """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Location_Category', namespaces))
        self.category = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Location_Category', namespaces))
        self.type = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Location_Subregion1', namespaces))
        self.subregion1 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Location_Subregion2', namespaces))
        self.subregion2 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Location_Subregion3', namespaces))
        self.subregion3 = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Detailed_Location', namespaces))
        self.detailed_location = util.testXMLValue(val)


class Data_Resolution(object):
    """ Process Data_Resolution"""
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Latitude_Resolution', namespaces))
        self.y = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Longitude_Resolution', namespaces))
        self.x = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Horizontal_Resolution_Range', namespaces))
        self.horizontal_res_range = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Vertical_Resolution', namespaces))
        self.vertical_res = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Vertical_Resolution_Range', namespaces))
        self.vertical_res_range = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Temporal_Resolution', namespaces))
        self.temporal_res = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Temporal_Resolution_Range', namespaces))
        self.temporal_res_range = util.testXMLValue(val)


class Data_Center(object):
    """ Process Data_Center """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Data_Center_Name', namespaces))
        self.name = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Data_Center_URL', namespaces))
        self.url = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Data_Set_ID', namespaces))
        self.data_set_id = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Personnel', namespaces))
        self.personnel = util.testXMLValue(val)


class Distribution(object):
    """ Process Distribution """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Distribution_Media', namespaces))
        self.media = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Distribution_Size', namespaces))
        self.size = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Distribution_Format', namespaces))
        self.format = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Fees', namespaces))
        self.fees = util.testXMLValue(val)


class Multimedia_Sample(object):
    """ Process Multimedia_Sample """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:File', namespaces))
        self.file = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:URL', namespaces))
        self.url = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Format', namespaces))
        self.format = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Caption', namespaces))
        self.caption = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Description', namespaces))
        self.description = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Visualization_URL', namespaces))
        self.vis_url = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Visualization_Type', namespaces))
        self.vis_type = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Visualization_Subtype', namespaces))
        self.vis_subtype = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Visualization_Duration', namespaces))
        self.vis_duration = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Visualization_File_Size', namespaces))
        self.file_size = util.testXMLValue(val)


class Related_URL(object):
    """ Process Related_URL """
    def __init__(self, md):
        self.content_type = []
        for el in md.findall(util.nspath_eval('dif:URL_Content_Type', namespaces)):
            self.content_type.append(URL_Content_Type(el))

        val = md.find(util.nspath_eval('dif:URL', namespaces))
        self.url = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:Description', namespaces))
        self.description = util.testXMLValue(val)


class URL_Content_Type(object):
    """ Process URL_Content_Type """
    def __init__(self, md):
        val = md.find(util.nspath_eval('dif:Type', namespaces))
        self.type = util.testXMLValue(val)

        val = md.find(util.nspath_eval('dif:SubType', namespaces))
        self.subtype = util.testXMLValue(val)



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2025 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

import codecs
from collections import OrderedDict
import copy
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from io import StringIO, BytesIO
import os
import re
import sys
from typing import Union
from urllib.parse import urlsplit, urlencode, urlparse, parse_qs, urlunparse, parse_qsl
import warnings

from dateutil import parser
import requests
from requests.auth import AuthBase

from owslib.etree import etree, ParseError
from owslib.namespaces import Namespaces

"""
Utility functions and classes
"""


class ServiceException(Exception):
    # TODO: this should go in ows common module when refactored.
    pass


# http://stackoverflow.com/questions/6256183/combine-two-dictionaries-of-dictionaries-python
def dict_union(d1, d2):
    return dict((x, (dict_union(d1.get(x, {}), d2[x]) if isinstance(d2.get(x), dict) else d2.get(x, d1.get(x))))
                for x in set(list(d1.keys()) + list(d2.keys())))


# Infinite DateTimes for Python.  Used in SWE 2.0 and other OGC specs as "INF" and "-INF"
class InfiniteDateTime(object):
    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return True

    def timetuple(self):
        return tuple()


class NegativeInfiniteDateTime(object):
    def __lt__(self, other):
        return True

    def __gt__(self, other):
        return False

    def timetuple(self):
        return tuple()


first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def format_string(prop_string):
    """
        Formats a property string to remove spaces and go from CamelCase to pep8
        from: http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case
    """
    if prop_string is None:
        return ''
    st_r = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', prop_string)
    st_r = st_r.replace(' ', '')
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', st_r).lower()


def xml_to_dict(root, prefix=None, depth=1, diction=None):
    """
        Recursively iterates through an xml element to convert each element in the tree to a (key,val).
        Where key is the element tag and val is the inner-text of the element.
        Note that this recursively go through the tree until the depth specified.

        Parameters
        ===========
        :root - root xml element, starting point of iteration
        :prefix - a string to prepend to the resulting key (optional)
        :depth - the number of depths to process in the tree (optional)
        :diction - the dictionary to insert the (tag,text) pairs into (optional)

        Return
        =======
        Dictionary of (key,value); where key is the element tag stripped of namespace and cleaned up to be pep8 and
        value is the inner-text of the element. Note that duplicate elements will be replaced by the last element of the
        same tag in the tree.
    """
    ret = diction if diction is not None else dict()
    for child in root:
        val = testXMLValue(child)
        # skip values that are empty or None
        if val is None or val == '':
            if depth > 1:
                ret = xml_to_dict(child, prefix=prefix, depth=(depth - 1), diction=ret)
            continue

        key = format_string(child.tag.split('}')[-1])

        if prefix is not None:
            key = prefix + key

        ret[key] = val
        if depth > 1:
            ret = xml_to_dict(child, prefix=prefix, depth=(depth - 1), diction=ret)

    return ret


class ResponseWrapper(object):
    """
    Return object type from openURL.

    Provides a thin shim around requests response object to maintain code compatibility.
    """
    def __init__(self, response):
        self._response = response

    def info(self):
        return self._response.headers

    def read(self):
        return self._response.content

    def geturl(self):
        return self._response.url.replace('&&', '&')

    # @TODO: __getattribute__ for poking at response


def openURL(url_base, data=None, method='Get', cookies=None, username=None, password=None, timeout=30, headers=None,
            verify=True, cert=None, auth=None):
    """
    Function to open URLs.

    Uses requests library but with additional checks for OGC service exceptions and url formatting.
    Also handles cookies and simple user password authentication.

    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param verify: (optional) whether the SSL cert will be verified. A CA_BUNDLE path can also be provided.
                   Defaults to ``True``.
    :param cert: (optional) A file with a client side certificate for SSL authentication
                 to send with the :class:`Request`.
    :param auth: Instance of owslib.util.Authentication
    """

    headers = headers if headers is not None else {}
    rkwargs = {}

    rkwargs['timeout'] = timeout

    if auth:
        if username:
            auth.username = username
        if password:
            auth.password = password
        if cert:
            auth.cert = cert
        verify = verify and auth.verify
    else:
        auth = Authentication(username, password, cert, verify)

    if auth.username and auth.password:
        rkwargs['auth'] = (auth.username, auth.password)
    elif auth.auth_delegate is not None:
        rkwargs['auth'] = auth.auth_delegate

    rkwargs['cert'] = auth.cert
    rkwargs['verify'] = verify

    # FIXUP for WFS in particular, remove xml style namespace
    # @TODO does this belong here?
    method = method.split("}")[-1]

    if method.lower() == 'post':
        try:
            etree.fromstring(data)
            headers['Content-Type'] = 'text/xml'
        except (ParseError, UnicodeEncodeError):
            pass

        rkwargs['data'] = data

    elif method.lower() == 'get':
        rkwargs['params'] = data

    else:
        raise ValueError("Unknown method ('%s'), expected 'get' or 'post'" % method)

    if cookies is not None:
        rkwargs['cookies'] = cookies

    req = requests.request(method.upper(), url_base, headers=headers, **rkwargs)

    if req.status_code in [400, 401, 403]:
        raise ServiceException(req.text)

    if req.status_code in [404, 500, 502, 503, 504]:    # add more if needed
        req.raise_for_status()

    # check for service exceptions without the http header set
    if 'Content-Type' in req.headers and \
            req.headers['Content-Type'] in ['text/xml', 'application/xml', 'application/vnd.ogc.se_xml']:
        # just in case 400 headers were not set, going to have to read the xml to see if it's an exception report.
        se_tree = etree.fromstring(req.content)

        # to handle the variety of namespaces and terms across services
        # and versions, especially for "legacy" responses like WMS 1.3.0
        possible_errors = [
            '{http://www.opengis.net/ows}Exception',
            '{http://www.opengis.net/ows/1.1}Exception',
            '{http://www.opengis.net/ogc}ServiceException',
            'ServiceException'
        ]

        for possible_error in possible_errors:
            serviceException = se_tree.find(possible_error)
            if serviceException is not None:
                # and we need to deal with some message nesting
                raise ServiceException('\n'.join([t.strip() for t in serviceException.itertext() if t.strip()]))

    return ResponseWrapper(req)


# default namespace for nspath is OWS common
OWS_NAMESPACE = 'http://www.opengis.net/ows/1.1'


def nspath(path, ns=OWS_NAMESPACE):

    """

    Prefix the given path with the given namespace identifier.

    Parameters
    ----------

    - path: ElementTree API Compatible path expression
    - ns: the XML namespace URI.

    """

    if ns is None or path is None:
        return -1

    components = []
    for component in path.split('/'):
        if component != '*':
            component = '{%s}%s' % (ns, component)
        components.append(component)
    return '/'.join(components)


def nspath_eval(xpath, namespaces):
    ''' Return an etree friendly xpath '''
    out = []
    for chunks in xpath.split('/'):
        namespace, element = chunks.split(':')
        out.append('{%s}%s' % (namespaces[namespace], element))
    return '/'.join(out)


def cleanup_namespaces(element):
    """ Remove unused namespaces from an element """
    etree.cleanup_namespaces(element)
    return element


def add_namespaces(root, ns_keys):
    if isinstance(ns_keys, str):
        ns_keys = [ns_keys]

    namespaces = Namespaces()

    ns_keys = [(x, namespaces.get_namespace(x)) for x in ns_keys]

    # lxml does not support setting xmlns attributes
    # Update the elements nsmap with new namespaces
    new_map = root.nsmap
    for key, link in ns_keys:
        if link is not None:
            new_map[key] = link
    # Recreate the root element with updated nsmap
    new_root = etree.Element(root.tag, nsmap=new_map)
    # Carry over attributes
    for a, v in list(root.items()):
        new_root.set(a, v)
    # Carry over children
    for child in root:
        new_root.append(deepcopy(child))
    return new_root


def getXMLInteger(elem, tag):
    """
    Return the text within the named tag as an integer.

    Raises an exception if the tag cannot be found or if its textual
    value cannot be converted to an integer.

    Parameters
    ----------

    - elem: the element to search within
    - tag: the name of the tag to look for

    """
    e = elem.find(tag)
    if e is None:
        raise ValueError('Missing %s in %s' % (tag, elem))
    return int(e.text.strip())


def getXMLTree(rsp: ResponseWrapper) -> etree:
    """
    Parse a response into an XML elementtree instance
    and raise a ValueError if the server returns a
    non-XML response. The response may contain a useful
    error message from the server.

    Parameters
    ----------

    @param rsp: the ResponseWrapper for the XML request
    """

    raw_text = strip_bom(rsp.read())
    et = etree.fromstring(raw_text)

    # check for response type - if it is not xml then raise an error
    # if the server doesn't provide a Content-Type then assume xml
    content_type = rsp.info().get('Content-Type', 'text/xml')
    url = rsp.geturl()

    has_xml_tag = raw_text.lstrip().startswith(b'<?xml')

    xml_types = ['text/xml', 'application/xml', 'application/vnd.ogc.wms_xml']
    if not any(xt in content_type.lower() for xt in xml_types) and not has_xml_tag:
        html_body = et.find('BODY')  # note this is case-sensitive
        if html_body is not None and len(html_body.text) > 0:
            response_text = html_body.text.strip("\n")
        else:
            response_text = raw_text

        raise ValueError("%s responded with Content-Type '%s': '%s'" %
                         (url, content_type, response_text))

    return et


def testXMLValue(val, attrib=False):
    """

    Test that the XML value exists, return val.text, else return None

    Parameters
    ----------

    - val: the value to be tested

    """

    if val is not None:
        if attrib:
            return val.strip()
        elif val.text:
            return val.text.strip()
        else:
            return None
    else:
        return None


def testXMLAttribute(element, attribute):
    """

    Test that the XML element and attribute exist, return attribute's value, else return None

    Parameters
    ----------

    - element: the element containing the attribute
    - attribute: the attribute name

    """
    if element is not None:
        return element.get(attribute)

    return None


def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, password=None, auth=None, headers=None):
    """

    Invoke an HTTP POST request

    Parameters
    ----------

    - url: the URL of the server
    - request: the request message
    - lang: the language
    - timeout: timeout in seconds
    - auth: owslib.util.Auth instance
    - headers: HTTP headers to send with requests

    """

    if url is None:
        raise ValueError("URL required")

    u = urlsplit(url)

    headers_ = {
        'User-Agent': 'OWSLib (https://owslib.readthedocs.io',
        'Content-type': 'text/xml',
        'Accept': 'text/xml,application/xml',
        'Accept-Language': lang,
        'Accept-Encoding': 'gzip,deflate',
        'Host': u.netloc,
    }

    if headers:
        headers_.update(headers)

    if isinstance(request, dict):
        headers_['Content-type'] = 'application/json'
        headers_.pop('Accept')

    rkwargs = {}

    if auth:
        if username:
            auth.username = username
        if password:
            auth.password = password
    else:
        auth = Authentication(username, password)
    if auth.username is not None and auth.password is not None:
        rkwargs['auth'] = (auth.username, auth.password)
    elif auth.auth_delegate is not None:
        rkwargs['auth'] = auth.auth_delegate
    rkwargs['verify'] = auth.verify
    rkwargs['cert'] = auth.cert

    if not isinstance(request, dict):
        return requests.post(url, request, headers=headers_, **rkwargs)
    else:
        return requests.post(url, json=request, headers=headers_, **rkwargs)


def http_prepare(*args, **kwargs):
    # Copy input kwargs so the dict can be modified
    rkwargs = copy.deepcopy(kwargs)

    # Use Authentication instance if provided, else create one
    auth = rkwargs.pop('auth', None)
    if auth is not None:
        if isinstance(auth, (tuple, list)):
            auth = Authentication(*auth)
    else:
        auth = Authentication()

    # Populate values with other arguments supplied
    if 'username' in rkwargs:
        auth.username = rkwargs.pop('username')
    if 'password' in rkwargs:
        auth.password = rkwargs.pop('password')
    if 'cert' in rkwargs:
        auth.cert = rkwargs.pop('cert')
    if 'verify' in rkwargs:
        auth.verify = rkwargs.pop('verify')

    # Build keyword args for call to requests.get()
    if auth.username and auth.password:
        rkwargs.setdefault('auth', (auth.username, auth.password))
    elif auth.auth_delegate is not None:
        rkwargs['auth'] = auth.auth_delegate
    else:
        rkwargs.setdefault('auth', None)
    rkwargs.setdefault('cert', rkwargs.get('cert'))
    rkwargs.setdefault('verify', rkwargs.get('verify', True))

    return rkwargs


def http_get(*args, **kwargs):
    rkwargs = http_prepare(*args, **kwargs)
    return requests.get(*args, **rkwargs)


def http_put(*args, **kwargs):
    rkwargs = http_prepare(*args, **kwargs)

    if 'data' in kwargs:
        if isinstance(kwargs['data'], dict):
            rkwargs['json'] = kwargs['data']
            rkwargs.pop('data')
        else:
            rkwargs['data'] = kwargs['data']

    return requests.put(*args, **rkwargs)


def http_delete(*args, **kwargs):
    rkwargs = http_prepare(*args, **kwargs)
    return requests.delete(*args, **rkwargs)


def element_to_string(element, encoding=None, xml_declaration=False):
    """
    Returns a string from a XML object

    Parameters
    ----------
    - element: etree Element
    - encoding (optional): encoding in string form. 'utf-8', 'ISO-8859-1', etc.
    - xml_declaration (optional): whether to include xml declaration

    """

    output = None

    if encoding is None:
        encoding = "ISO-8859-1"

    if xml_declaration:
        if encoding in ['unicode', 'utf-8']:
            output = '<?xml version="1.0" encoding="utf-8" standalone="no"?>\n{}'.format(
                etree.tostring(element, encoding='unicode'))
        else:
            output = etree.tostring(element, encoding=encoding, xml_declaration=True)
    else:
        output = etree.tostring(element)

    return output


def xml2string(xml):
    """

    Return a string of XML object

    Parameters
    ----------

    - xml: xml string

    """
    warnings.warn("DEPRECIATION WARNING!  You should now use the 'element_to_string' method \
                   The 'xml2string' method will be removed in a future version of OWSLib.")
    return '<?xml version="1.0" encoding="ISO-8859-1" standalone="no"?>\n' + xml


def xmlvalid(xml, xsd):
    """

    Test whether an XML document is valid

    Parameters
    ----------

    - xml: XML content
    - xsd: pointer to XML Schema (local file path or URL)

    """

    xsd1 = etree.parse(xsd)
    xsd2 = etree.XMLSchema(xsd1)

    doc = etree.parse(StringIO(xml))
    return xsd2.validate(doc)


def xmltag_split(tag):
    ''' Return XML element bare tag name (without prefix) '''
    try:
        return tag.split('}')[1]
    except Exception:
        return tag


def getNamespace(element):
    ''' Utility method to extract the namespace from an XML element tag encoded as {namespace}localname. '''
    if element.tag[0] == '{':
        return element.tag[1:].split("}")[0]
    else:
        return ""


def build_get_url(base_url, params, overwrite=False, doseq=False):
    ''' Utility function to build a full HTTP GET URL from the service base URL and a dictionary of HTTP parameters.

    TODO: handle parameters case-insensitive?

    @param overwrite: boolean flag to allow overwrite of parameters of the base_url (default: False)
    '''

    qs_base = []
    if base_url.find('?') != -1:
        qs_base = parse_qsl(base_url.split('?')[1])

    qs_params = []
    for key, value in list(params.items()):
        qs_params.append((key, value))

    qs = qs_add = []
    if overwrite is True:
        # all params and additional base
        qs = qs_params
        qs_add = qs_base
    else:
        # all base and additional params
        qs = qs_base
        qs_add = qs_params

    pars = [x[0] for x in qs]

    for key, value in qs_add:
        if key not in pars:
            qs.append((key, value))

    urlqs = urlencode(tuple(qs), doseq=doseq)
    return base_url.split('?')[0] + '?' + urlqs


def dump(obj, prefix=''):
    '''Utility function to print to standard output a generic object with all its attributes.'''

    return "{} {}.{} : {}".format(prefix, obj.__module__, obj.__class__.__name__, obj.__dict__)


def getTypedValue(data_type, value):
    '''Utility function to cast a string value to the appropriate XSD type. '''

    # If the default value is empty
    if value is None:
        return

    if data_type == 'boolean':
        return True if value.lower() == 'true' else False
    elif data_type == 'integer':
        return int(value)
    elif data_type == 'float':
        return float(value)
    elif data_type == 'string':
        return str(value)
    else:
        return value  # no type casting


def extract_time(element):
    ''' return a datetime object based on a gml text string

ex:
<gml:beginPosition>2006-07-27T21:10:00Z</gml:beginPosition>
<gml:endPosition indeterminatePosition="now"/>

If there happens to be a strange element with both attributes and text,
use the text.
ex: <gml:beginPosition indeterminatePosition="now">2006-07-27T21:10:00Z</gml:beginPosition>
Would be 2006-07-27T21:10:00Z, not 'now'

'''
    if element is None:
        return None

    try:
        dt = parser.parse(element.text)
    except Exception:
        att = testXMLValue(element.attrib.get('indeterminatePosition'), True)
        if att and att == 'now':
            dt = datetime.utcnow().replace(tzinfo=timezone.utc)
        else:
            dt = None
    return dt


def extract_xml_list(elements):
    """
Some people don't have separate tags for their keywords and separate them with
a newline. This will extract out all of the keywords correctly.
"""
    if elements is None:
        return []

    keywords = (re.split(r'[\n\r]+', f.text) for f in elements if f.text)
    flattened = (item.strip() for sublist in keywords for item in sublist)
    remove_blank = [_f for _f in flattened if _f]
    return remove_blank


def strip_bom(raw_text):
    """ return the raw (assumed) xml response without the BOM
    """
    boms = [
        # utf-8
        codecs.BOM_UTF8,
        # utf-16
        codecs.BOM,
        codecs.BOM_BE,
        codecs.BOM_LE,
        codecs.BOM_UTF16,
        codecs.BOM_UTF16_LE,
        codecs.BOM_UTF16_BE,
        # utf-32
        codecs.BOM_UTF32,
        codecs.BOM_UTF32_LE,
        codecs.BOM_UTF32_BE
    ]

    if isinstance(raw_text, bytes):
        for bom in boms:
            if raw_text.startswith(bom):
                return raw_text[len(bom):]
    return raw_text


def clean_ows_url(url):
    """
    clean an OWS URL of basic service elements

    source: https://stackoverflow.com/a/11640565
    """

    if url is None or not url.startswith('http'):
        return url

    filtered_kvp = {}
    basic_service_elements = ('service', 'version', 'request')

    parsed = urlparse(url)
    qd = parse_qs(parsed.query, keep_blank_values=True)

    for key, value in list(qd.items()):
        if key.lower() not in basic_service_elements:
            filtered_kvp[key] = value

    newurl = urlunparse([
        parsed.scheme,
        parsed.netloc,
        parsed.path,
        parsed.params,
        urlencode(filtered_kvp, doseq=True),
        parsed.fragment
    ])

    return newurl


def bind_url(url):
    """binds an HTTP GET query string endpiont"""
    if url.find('?') == -1:  # like http://host/wms
        binder = '?'

    # if like http://host/wms?foo=bar& or http://host/wms?foo=bar
    if url.find('=') != -1:
        if url.find('&', -1) != -1:  # like http://host/wms?foo=bar&
            binder = ''
        else:  # like http://host/wms?foo=bar
            binder = '&'

    # if like http://host/wms?foo
    if url.find('?') != -1:
        if url.find('?', -1) != -1:  # like http://host/wms?
            binder = ''
        elif url.find('&', -1) == -1:  # like http://host/wms?foo=bar
            binder = '&'
    return '%s%s' % (url, binder)


def findall(root, xpath, attribute_name=None, attribute_value=None):
    """Find elements recursively from given root element based on
    xpath and possibly given attribute

    :param root: Element root element where to start search
    :param xpath: xpath defintion, like {http://foo/bar/namespace}ElementName
    :param attribute_name: name of possible attribute of given element
    :param attribute_value: value of the attribute
    :return: list of elements or None
    """

    found_elements = []

    if attribute_name is not None and attribute_value is not None:
        xpath = '%s[@%s="%s"]' % (xpath, attribute_name, attribute_value)
    found_elements = root.findall('.//' + xpath)

    if found_elements == []:
        found_elements = None
    return found_elements


def datetime_from_iso(iso):
    """returns a datetime object from dates in the format 2001-07-01T00:00:00Z or 2001-07-01T00:00:00.000Z """
    try:
        iso_datetime = datetime.strptime(iso, "%Y-%m-%dT%H:%M:%SZ")
    except Exception:
        iso_datetime = datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%fZ")
    return iso_datetime


def datetime_from_ansi(ansi):
    """Converts an ansiDate (expressed as a number = the nuber of days since the datum origin of ansi)
    to a Python datetime object.
    """

    datumOrigin = datetime(1600, 12, 31, 0, 0, 0)

    return datumOrigin + timedelta(ansi)


def is_number(s):
    """simple helper to test if value is number as requests with numbers don't
    need quote marks
    """
    try:
        float(s)
        return True
    except ValueError:
        return False


def makeString(value):
    # using repr unconditionally breaks things in some circumstances if a
    # value is already a string
    if type(value) is not str:
        sval = repr(value)
    else:
        sval = value
    return sval


def param_list_to_url_string(param_list, param_name):
    """Converts list of tuples for certain WCS GetCoverage keyword arguments
    (subsets, resolutions, sizes) to a url-encoded string
    """
    string = ''
    for param in param_list:
        if len(param) > 2:
            if not is_number(param[1]):
                string += "&" + urlencode({param_name: param[0] + '("' + makeString(param[1]) + '","' + makeString(param[2]) + '")'})  # noqa
            else:
                string += "&" + urlencode({param_name: param[0] + "(" + makeString(param[1]) + "," + makeString(param[2]) + ")"})  # noqa
        else:
            if not is_number(param[1]):
                string += "&" + urlencode({param_name: param[0] + '("' + makeString(param[1]) + '")'})  # noqa
            else:
                string += "&" + urlencode({param_name: param[0] + "(" + makeString(param[1]) + ")"})  # noqa
    return string


def is_vector_grid(grid_elem):
    pass


class Authentication(object):

    _USERNAME = None
    _PASSWORD = None
    _AUTH_DELEGATE = None
    _CERT = None
    _VERIFY = None

    def __init__(self, username=None, password=None,
                 cert=None, verify=True, shared=False,
                 auth_delegate=None):
        '''
        :param str username=None: Username for basic authentication, None for
            unauthenticated access (or if using cert/verify)
        :param str password=None: Password for basic authentication, None for
            unauthenticated access (or if using cert/verify)
        :param cert=None: Either a str (path to a combined certificate/key) or
            tuple/list of paths (certificate, key). If supplied, the target
            files must exist.
        :param verify=True: Either a bool (verify SSL certificates, use system
            CA bundle) or str (path to a specific CA bundle). If a str, the
            target file must exist.
        :param bool shared=False: Set to True to make the values be class-level
            attributes (shared among instances where shared=True) instead of
            instance-level (shared=False, default)
        :param AuthBase auth_delegate=None: Instance of requests' AuthBase to
            allow arbitrary authentication schemes - mutually exclusive with
            username/password arguments.
        '''
        self.shared = shared
        self._username = username
        self._password = password
        self._cert = cert
        self._verify = verify
        self._auth_delegate = auth_delegate

        # Trigger the setters to validate the parameters. This couldn't be done directly
        # since some parameters are mutually exclusive.
        self.username = username
        self.password = password
        self.cert = cert
        self.verify = verify
        self.auth_delegate = auth_delegate

    @property
    def username(self):
        if self.shared:
            return self._USERNAME
        return self._username

    @username.setter
    def username(self, value):
        if value is not None:

            if not isinstance(value, str):
                raise TypeError('Value for "username" must be a str')

            if self.auth_delegate is not None:
                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
                                 ' but not both')

        if self.shared:
            self.__class__._USERNAME = value
        else:
            self._username = value

    @property
    def password(self):
        if self.shared:
            return self._PASSWORD
        return self._password

    @password.setter
    def password(self, value):
        if value is not None:

            if not isinstance(value, str):
                raise TypeError('Value for "password" must be a str')

            if self.auth_delegate is not None:
                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
                                 ' but not both')

        if self.shared:
            self.__class__._PASSWORD = value
        else:
            self._password = value

    @property
    def cert(self):
        if self.shared:
            return self._CERT
        return self._cert

    @cert.setter
    def cert(self, certificate, key=None):
        error = 'Value for "cert" must be a str path to a file or list/tuple of str paths'
        value = None
        if certificate is None:
            value = certificate
        elif isinstance(certificate, (list, tuple)):
            for _ in certificate:
                if not isinstance(_, str):
                    raise TypeError(error)
                os.stat(_)  # Raises OSError/FileNotFoundError if missing
            # Both paths supplied as same argument
            value = tuple(certificate)
        elif isinstance(certificate, str):
            os.stat(certificate)  # Raises OSError/FileNotFoundError if missing
            if isinstance(key, str):
                # Separate files for certificate and key
                value = (certificate, key)
            else:
                # Assume combined file of both certificate and key
                value = certificate
        else:
            raise TypeError(error)
        if self.shared:
            self.__class__._CERT = value
        else:
            self._cert = value

    @property
    def verify(self):
        if self.shared:
            return self._VERIFY
        return self._verify

    @verify.setter
    def verify(self, value):
        if value is None:
            pass  # Passthrough when clearing the value
        elif not isinstance(value, (bool, str)):
            raise TypeError(
                'Value for "verify" must a bool or str path to a file')
        elif isinstance(value, str):
            os.stat(value)  # Raises OSError/FileNotFoundError if missing
        if self.shared:
            self.__class__._VERIFY = value
        else:
            self._verify = value

    @property
    def auth_delegate(self):
        if self.shared:
            return self._AUTH_DELEGATE
        return self._auth_delegate

    @auth_delegate.setter
    def auth_delegate(self, value):
        if value is not None:
            if not isinstance(value, AuthBase):
                raise TypeError('Value for "auth_delegate" must be an instance of AuthBase')

            if self.username is not None or self.password is not None:
                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
                                 ' but not both')

        if self.shared:
            self.__class__._AUTH_DELEGATE = value
        else:
            self._auth_delegate = value

    @property
    def urlopen_kwargs(self):
        if self.auth_delegate is not None:
            raise NotImplementedError("The urlopen_kwargs property is not supported when auth_delegate is set")

        return {
            'username': self.username,
            'password': self.password,
            'cert': self.cert,
            'verify': self.verify
        }

    def __repr__(self, *args, **kwargs):
        return '<{} shared={} username={} password={} cert={} verify={} auth_delegate={}>'.format(
            self.__class__.__name__, self.shared, self.username, self.password, self.cert, self.verify,
            self.auth_delegate)


def str2bool(value: Union[bool, str]) -> bool:
    """
    helper function to return Python boolean
    type (source: https://stackoverflow.com/a/715468)

    :param value: value to be evaluated

    :returns: `bool` of whether the value is boolean-ish
    """

    value2 = False

    if isinstance(value, bool):
        value2 = value
    else:
        value2 = value.lower() in ('yes', 'true', 't', '1', 'on')

    return value2



# =============================================================================
# Copyright (c) 2020 Tom Kralidis
#
# Author: Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

import logging

from owslib.ogcapi.features import Features
from owslib.util import Authentication

LOGGER = logging.getLogger(__name__)


class Records(Features):
    """Abstraction for OGC API - Records"""

    def __init__(self, url: str, json_: str = None, timeout: int = 30,
                 headers: dict = None, auth: Authentication = None):
        __doc__ = Features.__doc__  # noqa
        super().__init__(url, json_, timeout, headers, auth)

    def records(self) -> list:
        """
        implements /collections filtered on records

        @returns: `list` of filtered collections object
        """

        records_ = []
        collections_ = super().collections()

        for c_ in collections_['collections']:
            if 'itemType' in c_ and c_['itemType'].lower() == 'record':
                records_.append(c_['id'])

        return records_
